/* filename: RanGene_no_gsl.c */
/* 15-Nov-2015
/******************************************************/
/* This gives pseudo-random numbers related to the    */
/* increment of Wiener processes and double           */
/* stochastic integrals.                              */
/******************************************************/

#include <stdio.h>

double genrand_real1(void);

int ran_gene_using_genrand_real1(int traject, int wdim,
				 double step, double base_step, double errC,
				 int MaxK, double work[], double rani[],
				 double tranij[], double ran_diag[],
				 int *rann_empty, double *rann_rest,
				 double init_acc_rani[]) {
  /* The polar method [Knuth:1981, p. 118] and the approximation method
     [Wiktorsson:2001] to the double stochastic integral */
  /* When an error occurs, this function will return a non-zero value. */

  /*
     Input arguments
     ----------------
     traject: number of trajectries,
     wdim: dimension of Wiener process,
     step: step length,
     base_step: a base step length to generete the increment of a normalized
     Wiener process, which means that dW(step)/sqrt(step)=
     dW1(base_step)/sqrt(step)+dW2(base_step)/sqrt(step)+...
     +dWn(base_step)/sqrt(step) for step=n*base_step,
     errC: error constant,
     MaxK: maximum number of terms for approximation to double stochastic
     integral,
     init_acc_rani: W/sqrt(h) at the initial time of the integral.

     Workspace arguments
     -------------------
     work: workspace of length (wdim+2*MaxK+2)*wdim+(wdim-1)*wdim/2.
      
     Output arguments
     ----------------
     rani: pointer of the head of wdim*traject RVs, in which the j-th
     element means delW/sqrt(h).
     tranij: pointer of the head of wdim*(wdim-1)*traject RVs, in which
     the j-th row has I_{1,j}/h, I_{2,j}/h, ..., I_{j-1,j}/h and 
     I_{j+1,j}/h, ..., I_{m,j}/h.
     tran_diag: pointer of the head of wdim*traject RVs, in which
     the j-th element means I_{j,j}/h.

     Input and output arguments
     --------------------------
     rann_empty: pointer of a flag if whose value is 1 a random number
     is set at *rann_rest, otherwise it empty,
     rann_rest: pointer of a random number.
  */

  static double pi2_6=1.6449340668482264, /* pi*pi/6 */
                pi2x24=236.87050562614461, /* 24*pi*pi */
                sq2=1.4142135623730950, /* sqrt(2)  */
                sq2pi=4.4428829381583662; /* sqrt(2)*pi */
    
  int ii, jj, kk, ll, ibase, wdimM1, loc_rann_empty, locMaxK, locMaxKM1,
    tranijDim, gdim, gdimM1, ibase_index, tmp_ibase_index, trajectM1,
    maxDiscardNum, tmpNum, tmpNumM1;
  double ranu1, ranu2, caps, loc_rann_rest, CBaseStep24PI2, rani_norm2, tmp, tmp2,
    locSqAn, tmpStep, sq_rate_step, eps, correction_term, rate_step;
  double *dummy, **X_dv_k, **Y, *tilX, **tilZ, *G, *xi_dv_sqstep,
    *loc_rani;

  if ((0.0>=base_step) || (step<base_step)) {
    printf("Error: the value of base_step is invalid in ran_gene_using_genrand_real1!\n");
    return 1;
  } else {
    rate_step=step/base_step;
    sq_rate_step=sqrt(rate_step);
    eps=0.1*base_step;
  }

  /* Preparation */
  trajectM1=traject-1;
  wdimM1=wdim-1;
  loc_rann_empty=*rann_empty;
  loc_rann_rest=*rann_rest;
  if (0!=wdimM1) {
    tranijDim=wdim*wdimM1;
    gdim=tranijDim/2;
    gdimM1=gdim-1;
    maxDiscardNum=MaxK*wdim+gdim;
    CBaseStep24PI2=errC*base_step*pi2x24;
    if (NULL == (X_dv_k = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in ranMulti_gene_using_genrand_real1\n");
      exit(0);
    }
    if (NULL == (Y = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in ranMulti_gene_using_genrand_real1\n");
      exit(0);
    }
    if (NULL == (tilZ = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in ranMulti_gene_using_genrand_real1\n");
      exit(0);
    }
    ii=0;
    tilX=&work[ii];
    ii+=wdim;
    for (jj=0; jj<wdim; jj++) {
      tilZ[jj]=&work[ii];
      ii+=wdimM1;
      X_dv_k[jj]=&work[ii];
      ii+=MaxK;
      Y[jj]=&work[ii];
      ii+=MaxK;
    }
    G=&work[ii];
    ii+=gdim;
    xi_dv_sqstep=&work[ii];
    ii+=wdim;
  }
  loc_rani=&work[ii];
  /*ii+=wdim;*/

  /* Initializing rani, ran_diag and tmpStep */
  for (ii=ibase=0;ii<traject;ii++) {
    for (jj=0; jj<wdim; jj++) {
      rani[ibase+jj]=0;
      ran_diag[ibase+jj]=0;
    }
    ibase+=wdim;
  }
  tmpStep=0.0;

  if (0!=wdimM1) {
    /* Initializing tranij */
    for (ii=ibase_index=0;ii<traject;ii++) {
      for (ll=0; ll<wdim; ll++) {
	tmp_ibase_index=ibase_index+ll*wdimM1;
	for (jj=0; jj<wdimM1; jj++) {
	  tranij[tmp_ibase_index+jj]=0;
	}
      }
      ibase_index+=tranijDim;
    }
  }

  while(eps<fabs(step-tmpStep)) {
    for (ii=ibase=ibase_index=0;ii<traject;ii++) {
      
      /* Generating delW/sqrt(h) and Iii/h */
      for (jj=0; jj<wdimM1; jj+=2) {
	while (1) {
	  ranu1 = 2*genrand_real1()-1;
	  ranu2 = 2*genrand_real1()-1;
	  caps = ranu1*ranu1 + ranu2*ranu2;
	  if ((1>caps) && (0<caps)) break;
	}
	caps = sqrt(-2.0*log(caps)/caps);
	if(1==loc_rann_empty) {
	  loc_rani[jj] = ranu1*caps;
	  loc_rani[jj+1] = ranu2*caps;
	} else {
	  loc_rani[jj] = loc_rann_rest;
	  loc_rani[jj+1] = ranu1*caps;
	  loc_rann_rest = ranu2*caps;
	  loc_rann_empty=0;
	}
	rani[ibase+jj]   += loc_rani[jj];
	rani[ibase+jj+1] += loc_rani[jj+1];
      }

      if(wdim!=jj) {
	if(0==loc_rann_empty) {
	  loc_rani[wdimM1] = loc_rann_rest;
	  loc_rann_empty = 1;
	} else {
	  while (1) {
	    ranu1 = 2*genrand_real1()-1;
	    ranu2 = 2*genrand_real1()-1;
	    caps = ranu1*ranu1 + ranu2*ranu2;
	    if ((1>caps) && (0<caps)) break;
	  }
	  caps = sqrt(-2.0*log(caps)/caps);
	  loc_rani[wdimM1] = ranu1*caps;
	  loc_rann_rest = ranu2*caps;
	  loc_rann_empty = 0;
	}
	rani[ibase+wdimM1] += loc_rani[wdimM1];
      }

      for (jj=0; jj<wdim; jj++) {
	ran_diag[ibase+jj]   +=
	  (init_acc_rani[ibase+jj]*sq_rate_step
	   +rani[ibase+jj]-loc_rani[jj])*loc_rani[jj]
	  +(loc_rani[jj]*loc_rani[jj]-1)/2.0;
      }

      if (0!=wdimM1) {
	/* Maximum of k in a traject */
	rani_norm2=0;
	for(jj=0; jj<wdim; jj++) {
	  rani_norm2+=loc_rani[jj]*loc_rani[jj];
	}

	tmp=sqrt(tranijDim*(wdim+4*rani_norm2)/CBaseStep24PI2);
	for(locMaxK=0; locMaxK<tmp; locMaxK++) {
	}
	if(MaxK<locMaxK) {
	  printf("Error: locMaxK is too large in ran_gene_using_genrand_real1!\n");
	  return 1;
	}

	locMaxKM1=locMaxK-1;

	/* Calculating sqrt(an) */
	locSqAn=0;
	for (kk=locMaxK; 0<kk; kk--) {
	  locSqAn+=1.0/(kk*kk);
	}
	
	locSqAn=sqrt(pi2_6-locSqAn);
	
	for (jj=0; jj<wdim; jj++) {
	  /* Generating X/k and tilX for jj */
	  for (kk=0; kk<locMaxKM1; kk+=2) {
	    while (1) {
	      ranu1 = 2*genrand_real1()-1;
	      ranu2 = 2*genrand_real1()-1;
	      caps = ranu1*ranu1 + ranu2*ranu2;
	      if ((1>caps) && (0<caps)) break;
	    }
	    caps = sqrt(-2.0*log(caps)/caps);
	    if(1==loc_rann_empty) {
	      X_dv_k[jj][kk] = ranu1*caps;
	      X_dv_k[jj][kk+1] = ranu2*caps;
	    } else {
	      X_dv_k[jj][kk] = loc_rann_rest;
	      X_dv_k[jj][kk+1] = ranu1*caps;
	      loc_rann_rest = ranu2*caps;
	      loc_rann_empty=0;
	    }
	  }
	  if(kk!=locMaxK) {
	    if(0==loc_rann_empty) {
	      X_dv_k[jj][locMaxKM1] = loc_rann_rest;
	      loc_rann_empty = 1;
	    } else {
	      while (1) {
		ranu1 = 2*genrand_real1()-1;
		ranu2 = 2*genrand_real1()-1;
		caps = ranu1*ranu1 + ranu2*ranu2;
		if ((1>caps) && (0<caps)) break;
	      }
	      caps = sqrt(-2.0*log(caps)/caps);
	      X_dv_k[jj][locMaxKM1] = ranu1*caps;
	      loc_rann_rest = ranu2*caps;
	      loc_rann_empty = 0;
	    }
	  }

	  for (kk=0; kk<locMaxK; kk++) {
	    X_dv_k[jj][kk]/=(1+kk);
	  }
	  tilX[jj]=0;
	  for (kk=locMaxKM1; 0<=kk; kk--) {
	    tilX[jj]+=X_dv_k[jj][kk];
	  }
	  /* Completed generating X/k and tilX for jj */
	  
	  /* Generating Y for jj */
	  for (kk=0; kk<locMaxKM1; kk+=2) {
	    while (1) {
	      ranu1 = 2*genrand_real1()-1;
	      ranu2 = 2*genrand_real1()-1;
	      caps = ranu1*ranu1 + ranu2*ranu2;
	      if ((1>caps) && (0<caps)) break;
	    }
	    caps = sqrt(-2.0*log(caps)/caps);
	    if(1==loc_rann_empty) {
	      Y[jj][kk] = ranu1*caps;
	      Y[jj][kk+1] = ranu2*caps;
	    } else {
	      Y[jj][kk] = loc_rann_rest;
	      Y[jj][kk+1] = ranu1*caps;
	      loc_rann_rest = ranu2*caps;
	      loc_rann_empty=0;
	    }
	  }
	  if(kk!=locMaxK) {
	    if(0==loc_rann_empty) {
	      Y[jj][locMaxKM1] = loc_rann_rest;
	      loc_rann_empty = 1;
	    } else {
	      while (1) {
		ranu1 = 2*genrand_real1()-1;
		ranu2 = 2*genrand_real1()-1;
		caps = ranu1*ranu1 + ranu2*ranu2;
		if ((1>caps) && (0<caps)) break;
	      }
	      caps = sqrt(-2.0*log(caps)/caps);
	      Y[jj][locMaxKM1] = ranu1*caps;
	      loc_rann_rest = ranu2*caps;
	      loc_rann_empty = 0;
	    }
	  }
	  /* Completed generating Y for jj */
	  
	} /* End of the loop for jj */
	
	/* Generating G */
	for (jj=0; jj<gdimM1; jj+=2) {
	  while (1) {
	    ranu1 = 2*genrand_real1()-1;
	    ranu2 = 2*genrand_real1()-1;
	    caps = ranu1*ranu1 + ranu2*ranu2;
	    if ((1>caps) && (0<caps)) break;
	  }
	  caps = sqrt(-2.0*log(caps)/caps);
	  if(1==loc_rann_empty) {
	    G[jj] = ranu1*caps;
	    G[jj+1] = ranu2*caps;
	  } else {
	    G[jj] = loc_rann_rest;
	    G[jj+1] = ranu1*caps;
	    loc_rann_rest = ranu2*caps;
	    loc_rann_empty=0;
	  }
	} /* End of the loop for jj */
	if(jj!=gdim) {
	  if(0==loc_rann_empty) {
	    G[gdimM1] = loc_rann_rest;
	    loc_rann_empty = 1;
	  } else {
	    while (1) {
	      ranu1 = 2*genrand_real1()-1;
	      ranu2 = 2*genrand_real1()-1;
	      caps = ranu1*ranu1 + ranu2*ranu2;
	      if ((1>caps) && (0<caps)) break;
	    }
	    caps = sqrt(-2.0*log(caps)/caps);
	    G[gdimM1] = ranu1*caps;
	    loc_rann_rest = ranu2*caps;
	    loc_rann_empty = 0;
	  }
	}
	/* Completed generatig G */
	
	/* Calculating tilZ */
	for (jj=0; jj<wdim; jj++) {
	  for (ll=0; ll<jj; ll++) {
	    tilZ[jj][ll]=0;
	    for (kk=0; kk<locMaxK; kk++) {
	      tilZ[jj][ll]+=X_dv_k[jj][kk]*Y[ll][kk];
	    }
	  }
	  for (ll=jj+1; ll<wdim; ll++) {
	    tilZ[jj][ll-1]=0;
	    for (kk=0; kk<locMaxK; kk++) {
	      tilZ[jj][ll-1]+=X_dv_k[jj][kk]*Y[ll][kk];
	    }
	  }
	}
	/* Completed calculating tilZ */
	
	/* Calculating xi_dv_sqstep */
	for (jj=0; jj<wdim; jj++) {
	  xi_dv_sqstep[jj]=0;
	  for (ll=jj+1; ll<wdim; ll++) {
	    kk=jj*wdim-jj*(jj+1)/2+ll-jj-1;
	    xi_dv_sqstep[jj]+=loc_rani[ll]*G[kk];
	  }
	  for (ll=0; ll<jj; ll++) {
	    kk=ll*wdim-ll*(ll+1)/2+jj-ll-1;
	    xi_dv_sqstep[jj]-=loc_rani[ll]*G[kk];
	  }
	}
	
	/* Calculating tilA/h */
	tmp=1+sqrt(1+rani_norm2);
	for (jj=0; jj<wdimM1; jj++) {
	  for (ll=jj+1; ll<wdim; ll++) {
	    kk=jj*wdim-jj*(jj+1)/2+ll-jj-1;
	    correction_term
	      =locSqAn*(G[kk]+(loc_rani[ll]*xi_dv_sqstep[jj]
			       -loc_rani[jj]*xi_dv_sqstep[ll])/tmp);
	    tilZ[jj][ll-1]
	      =((tilZ[jj][ll-1]-tilZ[ll][jj])/sq2
		+tilX[jj]*loc_rani[ll]-tilX[ll]*loc_rani[jj]
		+correction_term
		)/sq2pi;
	    tilZ[ll][jj]=-tilZ[jj][ll-1];
	  }
	}
	/* Completed calculating tilA/h */
	
	/* Calculating Transpose(Iij)/h */
	for (ll=0; ll<wdimM1; ll++) {
	  tmp_ibase_index=ibase_index+ll*wdimM1;
	  for (jj=ll+1; jj<wdim; jj++) {
	    tranij[tmp_ibase_index+jj-1]+=
	      (init_acc_rani[ibase+jj]*sq_rate_step
	       +rani[ibase+jj]-loc_rani[jj])*loc_rani[ll]
	      +loc_rani[jj]*loc_rani[ll]/2+tilZ[jj][ll];
	  }
	}
	for (ll=1; ll<wdim; ll++) {
	  tmp_ibase_index=ibase_index+ll*wdimM1;
	  for (jj=0; jj<ll; jj++) {
	    tranij[tmp_ibase_index+jj]+=
	      (init_acc_rani[ibase+jj]*sq_rate_step
	       +rani[ibase+jj]-loc_rani[jj])*loc_rani[ll]
	      +loc_rani[jj]*loc_rani[ll]/2+tilZ[jj][ll-1];
	  }
	}

	if (eps>=fabs(step-(tmpStep+base_step))) {
	  for (ll=0; ll<wdimM1; ll++) {
	    tmp_ibase_index=ibase_index+ll*wdimM1;
	    for (jj=ll+1; jj<wdim; jj++) {
	      tranij[tmp_ibase_index+jj-1]=
		(tranij[tmp_ibase_index+jj-1]
		 -init_acc_rani[ibase+jj]*sq_rate_step*rani[ibase+ll]
		 )/rate_step;
	    }
	  }
	  for (ll=1; ll<wdim; ll++) {
	    tmp_ibase_index=ibase_index+ll*wdimM1;
	    for (jj=0; jj<ll; jj++) {
	      tranij[tmp_ibase_index+jj]=
		(tranij[tmp_ibase_index+jj]
		 -init_acc_rani[ibase+jj]*sq_rate_step*rani[ibase+ll]
		 )/rate_step;
	    }
	  }
	/* Completed calculating Transpose(Iij)/h */
	}
	
	ibase_index+=tranijDim;
      } /* End of if for 0!=wdimM1 */

      if (eps>=fabs(step-(tmpStep+base_step))) {
	for (jj=0; jj<wdim; jj++) {
	  ran_diag[ibase+jj]=
	    (ran_diag[ibase+jj]
	     -init_acc_rani[ibase+jj]*sq_rate_step*rani[ibase+jj])/rate_step;
	  rani[ibase+jj]/=sq_rate_step;
	}
	/* Completed generating delW/sqrt(h) and Iii/h. */
      }
      
      ibase+=wdim;
    } /* End of the loop for ii. */
    
    tmpStep += base_step; /* Update on tmpStep */
  } /* End of the while for tmpStep */

  if (0!=wdimM1) {
    free(X_dv_k);
    free(Y);
    free(tilZ);
  }

  *rann_empty=loc_rann_empty;
  *rann_rest=loc_rann_rest;
  return 0;
}
