/* filename: SROCK.c */
/* 16-Nov-2015 */
/*
  In this file, the SROCKD1 and SROCKD2 methods proposed in
  [Komori:2013] were implemented for Ito SDEs. The SROCKD1
  and SROCKD2 methods correspond to the function
  "OMP_so1_srockD1_for_NonCommSDEs_WinMulti" and
  "OMP_so1_srock_for_NonCommSDEs_WinMulti", respectively.
  For the use of them, see the sample program "main_samp.c".

  PLEASE CITE THE FOLLOWING PAPER WHEN USING THESE:

  [1] Y. Komori and K. Burrage
      Strong first order S-ROCK methods for stochastic
      differential equations
      JCAM, 242 (2013), 261-274.

 */

#include <stdio.h>
#include <math.h>

#define MaxStageNum 104 /* Maximum stage number */

#define MaxCoreNum 8 /* 8, 1 *//* Maximum number of multi-core. */
/*
  In this program code, OpenMP is available. If you want to use it,
  remove a comment out in the line "#pragma omp parallel for".
 */

int GetSRKVal(int ss, double Mu[], double Ka[], double *sig, double *coe);

int GetSROCKD1Val(int ss, double coe[]);

extern int OMP_so1_srock_for_NonCommSDEs_WinMulti(int ydim,
						  unsigned long traj,
						  double *yvec,
						  double step,
						  double *ranni,
						  double *TransRanIij,
						  double *ranDiag,
						  void (*ffunc)(),
						  void (*gfunc_gene)(),
						  int wdim, int ss,
						  double work[], double *ynew)
/* This function performs strong first order SRK scheme based on rock.
   It gives all trajectries for one step concerning SDEs with
   a multi-dimentional Winer process. If an error occurs, it will
   return 1, otherwise 0.

   Input arguments
   ----------------
   ydim: dimension of SDEs,
   traj: number of trajectries,
   yvec: pointer of the head of all initial values, which are in order
         like yvec[0], yvec[1], ..., yvec[ydim-1] for the 1st trajectry,
	 yvec[ydim], yvec[ydim+1], ..., yvec[2*ydim-1] for the 2nd trajectry.
   step: step length,
   ranni: pointer of the head of wdim*traject RVs, in which the j-th
   element means delW/sqrt(h),
   TransRanIij: pointer of the head of wdim*(wdim-1)*traj RVs, in which
   the j-th row has I_{1,j}/h, I_{2,j}/h, ..., I_{j-1,j}/h and 
   I_{j+1,j}/h, ..., I_{m,j}/h,
   ranDiag: pointer of the head of wdim*traject RVs, in which the j-th element
   means I_{j,j}/h.
   ffunc: drift coefficient,
   gfunc_gene: diffusion coefficients,
   wdim: dimension of Wiener process.
   ss: stage numer of SRK scheme.

   Workspace arguments
   -------------------
   work: workspace of length MaxCoreNum*(8+3*wdim)*ydim.

   Output arguments
   ----------------
   ynew: pointer of the head of all solutions for one step, which are
         in a similar order to yvec.
*/
{
  static int static_flag=0;
  static unsigned long traj_mini;
  int errflag, sM2, sM3, wdimM1, wdimWdimM1, ii_par;
  /* See Research_for_Stability_func_of_S_stages_Chev_Strong_SRK_for_Ito.nb. */
  static double
    /* The coefficients in the stochastic part */
    B1sM1sM2=1.0,
    B1ssM2,
    /***/
    bt1sM2=1.0,
    /***/
    bt2sM1, bt2s;

  double Mu[MaxStageNum-2], Ka[MaxStageNum-3], sigma, coeff, sqstep;
  
  if((ydim <=0) || (step <= 0))
    {printf("negative argument err\n");return 1;}/* Check for arguments */
  sqstep = sqrt(step);
  
  wdimM1=wdim-1;
  wdimWdimM1=wdim*wdimM1;

  errflag=GetSRKVal(ss, Mu, Ka, &sigma, &coeff);
  if (1==errflag) {
    printf("Error: ss is not among our selection numbers!\n");
    printf("       it must satisfy 4<=ss<=22, or it must be\n");
    printf("       24, 26, 28, 30, 32, 35, 38, 41, 45, 49,\n");
    printf("       53, 58, 63, 68, 74, 80, 87, 95 or 104.\n");
    exit(1);
  }
  sM2=ss-2;
  sM3=ss-3;

  B1ssM2=-B1sM1sM2;
  bt2sM1=1.0/(2*B1sM1sM2);
  bt2s=-bt2sM1;

  if (0==static_flag) {
    if (0!=traj%MaxCoreNum) {
      printf("Error: Number of trajects must be %d multiple!", MaxCoreNum);
      exit(1);
    }
    traj_mini=traj/MaxCoreNum;
    static_flag=1;
  }

/*#pragma omp parallel for */
  for (ii_par=0; ii_par<MaxCoreNum; ii_par++) {
    unsigned long itr;
    int ii, jj, ll, ibaseRanni, ibaseTRanIij, ibase, ibase_work,
      ibase_work_step, ibase_index;
    double *yn, *yn1, *fn, *gn, *pjM2, *pjM1, *pj, *hgH0sM1, *hgH0s,
      tmp1, tmp2, tmp3;
    double *dummy, **sqhgHjsM2, **sqhgHjsM1, **sqhgHjs;
    double *wj, *TraIij, *I_diag;

    if (NULL == (sqhgHjsM2 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (sqhgHjsM1 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (sqhgHjs = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }

    ibase_work_step=8+3*wdim;
    ibase_work=(ibase_work_step*ydim)*ii_par; /**/
    ii=0;
    yn=&work[ibase_work+ii];
    ii+=ydim;
    fn=&work[ibase_work+ii];
    ii+=ydim;
    gn=&work[ibase_work+ii];
    ii+=ydim;
    pjM2=&work[ibase_work+ii];
    ii+=ydim;
    pjM1=&work[ibase_work+ii];
    ii+=ydim;
    pj=&work[ibase_work+ii];
    ii+=ydim;
    hgH0sM1=&work[ibase_work+ii];
    ii+=ydim;
    hgH0s=&work[ibase_work+ii];
    ii+=ydim;
    for (jj=0; jj<wdim; jj++) {
      sqhgHjsM2[jj]=&work[ibase_work+ii];
      ii+=ydim;
      sqhgHjsM1[jj]=&work[ibase_work+ii];
      ii+=ydim;
      sqhgHjs[jj]=&work[ibase_work+ii];
      ii+=ydim;
    }

    ibase=(ydim*traj_mini)*ii_par;
    ibaseRanni=(wdim*traj_mini)*ii_par;
    ibaseTRanIij=(wdimWdimM1*traj_mini)*ii_par;
    for(itr=0; itr<traj_mini; itr++) {
      wj=&ranni[ibaseRanni];
      I_diag=&ranDiag[ibaseRanni];
      ibaseRanni+=wdim;
      TraIij=&TransRanIij[ibaseTRanIij];
      ibaseTRanIij+=wdimWdimM1;
      yn1=&ynew[ibase];
      for(ii=0; ii<ydim; ii++) {
	yn[ii]=yvec[ibase+ii];
      }
      ibase+=ydim;

      /* See (9), (13) and (18) in [Komori:2013] */
      ffunc(yn,fn);
      tmp1=step*Mu[0];
      for(ii=0; ii<ydim; ii++) {
	pjM2[ii]=yn[ii];
	pjM1[ii]=yn[ii]+tmp1*fn[ii];
      }
      for(jj=1; jj<sM2; jj++) {
	ffunc(pjM1,fn);
	tmp1=step*Mu[jj];
	tmp2=1+Ka[jj-1];
	tmp3=-Ka[jj-1];
	for (ii=0; ii<ydim; ii++) {
	  pj[ii]=tmp1*fn[ii]+tmp2*pjM1[ii]+tmp3*pjM2[ii];
	  if (jj<sM3) {
	    pjM2[ii]=pjM1[ii];
	    pjM1[ii]=pj[ii];
	  }
	}
      }
      if (3==ss) {
	for(ii=0; ii<ydim; ii++) {
	  pj[ii]=pjM1[ii];
	}
      }

      for (jj=0; jj<wdim; jj++) {
	gfunc_gene(pj,jj,gn);
	for (ii=0; ii<ydim; ii++) {
	  sqhgHjsM2[jj][ii]=sqstep*gn[ii]; /* completed */
	}
      }

      ffunc(pj,fn);
      for(ii=0; ii<ydim; ii++) {
	hgH0sM1[ii]=step*fn[ii]; /* completed */
      }

      for (jj=0; jj<wdim; jj++) {
	ibase_index=jj*wdimM1;
	for(ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  for (ll=0; ll<jj; ll++) {
	    tmp1+=TraIij[ibase_index+ll]*sqhgHjsM2[ll][ii];
	  }
	  tmp1+=I_diag[jj]*sqhgHjsM2[jj][ii];
	  for (ll=jj; ll<wdimM1; ll++) {
	    tmp1+=TraIij[ibase_index+ll]*sqhgHjsM2[ll+1][ii];
	  }
	  sqhgHjsM1[jj][ii]=pj[ii]+B1sM1sM2*tmp1;
	  sqhgHjs[jj][ii]=pj[ii]+B1ssM2*tmp1;
	}
      }
      
      for (jj=0; jj<wdim; jj++) {
	gfunc_gene(sqhgHjsM1[jj],jj,gn);
	for (ii=0; ii<ydim; ii++) {
	  sqhgHjsM1[jj][ii]=sqstep*gn[ii]; /* completed */
	}
	gfunc_gene(sqhgHjs[jj],jj,gn);
	for (ii=0; ii<ydim; ii++) {
	  sqhgHjs[jj][ii]=sqstep*gn[ii]; /* completed */
	}
      }

      for (ii=0; ii<ydim; ii++) {
	hgH0s[ii]=pj[ii]+sigma*hgH0sM1[ii];
      }
      ffunc(hgH0s,fn);
      for(ii=0; ii<ydim; ii++) {
	hgH0s[ii]=step*fn[ii]; /* completed */
      }

      for (ii=0; ii<ydim; ii++) {
	tmp1=tmp2=tmp3=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=wj[jj]*sqhgHjsM2[jj][ii];
	  tmp2+=sqhgHjsM1[jj][ii];
	  tmp3+=sqhgHjs[jj][ii];
	}
	yn1[ii]=pj[ii]+sigma*(hgH0sM1[ii]+hgH0s[ii])
	  +coeff*(hgH0s[ii]-hgH0sM1[ii])+bt1sM2*tmp1+bt2sM1*tmp2+bt2s*tmp3;
      }
      
    } /* End of loop for itr */
    free(sqhgHjsM2);
    free(sqhgHjsM1);
    free(sqhgHjs);
  }
  return 0;
}

extern int OMP_so1_srockD1_for_NonCommSDEs_WinMulti(int ydim,
						    unsigned long traj,
						    double *yvec,
						    double step,
						    double *ranni,
						    double *TransRanIij,
						    double *ranDiag,
						    void (*ffunc)(),
						    void (*gfunc_gene)(),
						    int wdim, int ss,
						    double work[], double *ynew)
/* This function performs strong first order SRK scheme with deterministic
   order one based on rock.
   It gives all trajectries for one step concerning SDEs with
   a multi-dimentional Winer process. If an error occurs, it will
   return 1, otherwise 0.

   Input arguments
   ----------------
   ydim: dimension of SDEs,
   traj: number of trajectries,
   yvec: pointer of the head of all initial values, which are in order
         like yvec[0], yvec[1], ..., yvec[ydim-1] for the 1st trajectry,
	 yvec[ydim], yvec[ydim+1], ..., yvec[2*ydim-1] for the 2nd trajectry.
   step: step length,
   ranni: pointer of the head of wdim*traject RVs, in which the j-th
   element means delW/sqrt(h),
   TransRanIij: pointer of the head of wdim*(wdim-1)*traj RVs, in which
   the j-th row has I_{1,j}/h, I_{2,j}/h, ..., I_{j-1,j}/h and 
   I_{j+1,j}/h, ..., I_{m,j}/h,
   ranDiag: pointer of the head of wdim*traject RVs, in which the j-th element
   means I_{j,j}/h.
   ffunc: drift coefficient,
   gfunc_gene: diffusion coefficients,
   wdim: dimension of Wiener process.
   ss: stage numer of SRK scheme.

   Workspace arguments
   -------------------
   work: workspace of length MaxCoreNum*(6+3*wdim)*ydim.

   Output arguments
   ----------------
   ynew: pointer of the head of all solutions for one step, which are
         in a similar order to yvec.
*/
{
  static int static_flag=0;
  static unsigned long traj_mini;
  int errflag, sM1, wdimM1, wdimWdimM1, ii_par;
  /* See Research_Part_B_for_Stability_func_of_S_stages_Chev_Strong_SRK_for_Ito.nb. */
  static double
    /* The coefficients in the stochastic part */
    B1sM1sM2=1.0,
    B1ssM2,
    /***/
    bt1sM2=1.0,
    /***/
    bt2sM1, bt2s;

  double coe[2*(MaxStageNum-1)+1], sqstep;
  
  if((ydim <=0) || (step <= 0))
    {printf("negative argument err\n");return 1;}/* Check for arguments */
  sqstep = sqrt(step);
  
  wdimM1=wdim-1;
  wdimWdimM1=wdim*wdimM1;

  errflag=GetSROCKD1Val(ss, coe);
  if (1==errflag) {
    printf("Error: ss is not among our selection numbers!\n");
    printf("       it must satisfy 4<=ss<=22, or it must be\n");
    printf("       24, 26, 28, 30, 32, 35, 38, 41, 45, 49,\n");
    printf("       53, 58, 63, 68, 74, 80, 87, 95 or 104.\n");
    exit(1);
  }
  sM1=ss-1;

  B1ssM2=-B1sM1sM2;
  bt2sM1=1.0/(2*B1sM1sM2);
  bt2s=-bt2sM1;

  if (0==static_flag) {
    if (0!=traj%MaxCoreNum) {
      printf("Error: Number of trajects must be %d multiple!", MaxCoreNum);
      exit(1);
    }
    traj_mini=traj/MaxCoreNum;
    static_flag=1;
  }

/*#pragma omp parallel for */
  for (ii_par=0; ii_par<MaxCoreNum; ii_par++) {
    unsigned long itr;
    int ii, jj, ll, ibaseRanni, ibaseTRanIij, ibase, ibaseCoe, ibase_work,
      ibase_work_step, ibase_index;
    double *yn, *yn1, *fn, *gn, *kjM2, *kjM1, *kj, tmp1, tmp2, tmp3;
    double *dummy, **sqhgHjsM2, **sqhgHjsM1, **sqhgHjs;
    double *wj, *TraIij, *I_diag;

    if (NULL == (sqhgHjsM2 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (sqhgHjsM1 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (sqhgHjs = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }

    ibase_work_step=6+3*wdim;
    ibase_work=(ibase_work_step*ydim)*ii_par; /**/
    ii=0;
    yn=&work[ibase_work+ii];
    ii+=ydim;
    fn=&work[ibase_work+ii];
    ii+=ydim;
    gn=&work[ibase_work+ii];
    ii+=ydim;
    kjM2=&work[ibase_work+ii];
    ii+=ydim;
    kjM1=&work[ibase_work+ii];
    ii+=ydim;
    kj=&work[ibase_work+ii];
    ii+=ydim;
    for (jj=0; jj<wdim; jj++) {
      sqhgHjsM2[jj]=&work[ibase_work+ii];
      ii+=ydim;
      sqhgHjsM1[jj]=&work[ibase_work+ii];
      ii+=ydim;
      sqhgHjs[jj]=&work[ibase_work+ii];
      ii+=ydim;
    }

    ibase=(ydim*traj_mini)*ii_par;
    ibaseRanni=(wdim*traj_mini)*ii_par;
    ibaseTRanIij=(wdimWdimM1*traj_mini)*ii_par;
    for(itr=0; itr<traj_mini; itr++) {
      wj=&ranni[ibaseRanni];
      I_diag=&ranDiag[ibaseRanni];
      ibaseRanni+=wdim;
      TraIij=&TransRanIij[ibaseTRanIij];
      ibaseTRanIij+=wdimWdimM1;
      yn1=&ynew[ibase];
      for(ii=0; ii<ydim; ii++) {
	yn[ii]=yvec[ibase+ii];
      }
      ibase+=ydim;

      /* See (8), (13) and (18) in [Komori:2013]. */
      /* Calculations for K_{s} */
      ffunc(yn,fn);
      tmp1=step*coe[0];
      ibaseCoe=0;
      for(ii=0; ii<ydim; ii++) {
	kjM2[ii]=yn[ii];
	kjM1[ii]=yn[ii]+tmp1*fn[ii];
      }
      for(jj=1; jj<ss; jj++) {
	ffunc(kjM1,fn);
	tmp1=step*coe[ibaseCoe+1];
	tmp3=coe[ibaseCoe+2];
	tmp2=1-tmp3;
	ibaseCoe+=2;
	for (ii=0; ii<ydim; ii++) {
	  kj[ii]=tmp1*fn[ii]+tmp2*kjM1[ii]+tmp3*kjM2[ii];
	  if (jj<sM1) {
	    kjM2[ii]=kjM1[ii];
	    kjM1[ii]=kj[ii];
	  }
	}
      }

      for (jj=0; jj<wdim; jj++) {
	gfunc_gene(kj,jj,gn);
	for (ii=0; ii<ydim; ii++) {
	  sqhgHjsM2[jj][ii]=sqstep*gn[ii]; /* completed */
	}
      }

      for (jj=0; jj<wdim; jj++) {
	ibase_index=jj*wdimM1;
	for(ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  for (ll=0; ll<jj; ll++) {
	    tmp1+=TraIij[ibase_index+ll]*sqhgHjsM2[ll][ii];
	  }
	  tmp1+=I_diag[jj]*sqhgHjsM2[jj][ii];
	  for (ll=jj; ll<wdimM1; ll++) {
	    tmp1+=TraIij[ibase_index+ll]*sqhgHjsM2[ll+1][ii];
	  }
	  sqhgHjsM1[jj][ii]=kj[ii]+B1sM1sM2*tmp1;
	  sqhgHjs[jj][ii]=kj[ii]+B1ssM2*tmp1;
	}
      }
      
      for (jj=0; jj<wdim; jj++) {
	gfunc_gene(sqhgHjsM1[jj],jj,gn);
	for (ii=0; ii<ydim; ii++) {
	  sqhgHjsM1[jj][ii]=sqstep*gn[ii]; /* completed */
	}
	gfunc_gene(sqhgHjs[jj],jj,gn);
	for (ii=0; ii<ydim; ii++) {
	  sqhgHjs[jj][ii]=sqstep*gn[ii]; /* completed */
	}
      }

      for (ii=0; ii<ydim; ii++) {
	tmp1=tmp2=tmp3=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=wj[jj]*sqhgHjsM2[jj][ii];
	  tmp2+=sqhgHjsM1[jj][ii];
	  tmp3+=sqhgHjs[jj][ii];
	}
	yn1[ii]=kj[ii]+bt1sM2*tmp1+bt2sM1*tmp2+bt2s*tmp3;
      }
      
    } /* End of loop for itr */
    free(sqhgHjsM2);
    free(sqhgHjsM1);
    free(sqhgHjs);
  }
  return 0;
}
