/* Filename: main_samp.c */
/* (17-Nov-2015) */

/******************************************************/
/* This program gives an example to solve Ito SDEs by */
/* SROCKD1 or SROCKD2. The example is a               */
/* two-dimensional Ito SDE with two Wiener processes: */
/*    dy=ffunc dt + sum_{j=1}^{2} gfuncj dwj.         */
/******************************************************/

#include <stdio.h>
#include <math.h>
#include <direct.h>

/* The following flag is used to assign an srk method. */
#define SRK_TYPE 1 /* 1, 2 */

#if (1 == SRK_TYPE)
#  define SRK    OMP_so1_srock_for_NonCommSDEs_WinMulti
#elif (2 == SRK_TYPE)
#  define SRK    OMP_so1_srockD1_for_NonCommSDEs_WinMulti
#else
#  define SRK    /* nothing */
#endif

/* Maximum of an independent variable x. */
#define  T_END               1.0

/* Minimum of a step size. */
#define  STEP_SIZE        1.0/32.0
#define  FINE_STEP_SIZE   STEP_SIZE /* For another example, STEP_SIZE/4.0 */

#  define  YDIM         2
#  define  WDIM         2

#define  MaxK    256 /* 64, 256, 512 */

#define  SQ2 1.4142135623730950 /* sqrt(2) */
#define  A11      -273.0/512
#define  A21      -1.0/160
#define  A22      (-785.0/512+SQ2/8)
#define  B11       1.0/4
#define  B12       1.0/16
#define  B21       (1-2*SQ2)/4
#define  B221      1.0/10
#define  B222      1.0/16

#define  TRAJECT        1000UL

#define  NWinMax         WDIM*TRAJECT
#define  NWinWinM1Max    WDIM*(WDIM-1)*TRAJECT
#define  NArrayMax       YDIM*TRAJECT

/* ffunc is a drift coefficient.  */
static void ffunc(double ynvec[],double foutput[]);

/* gfunc_gene is diffusion coefficients. */
static void gfunc_gene(double ynvec[],int i_th,double goutput[]);

/* Prallelized Stochastic Runge-Kutta Multi Win Any */
extern int OMP_so1_srock_for_NonCommSDEs_WinMulti(int ydim,
						  unsigned long traj,
						  double *yvec,
						  double step,
						  double *ranni,
						  double *TransRanIij,
						  double *ranDiag,
						  void (*ffunc)(),
						  void (*gfunc_gene)(),
						  int wdim, int ss,
						  double work[], double *ynew);

extern int OMP_so1_srockD1_for_NonCommSDEs_WinMulti(int ydim,
						    unsigned long traj,
						    double *yvec,
						    double step,
						    double *ranni,
						    double *TransRanIij,
						    double *ranDiag,
						    void (*ffunc)(),
						    void (*gfunc_gene)(),
						    int wdim, int ss,
						    double work[],
						    double *ynew);

void init_genrand(unsigned long s);

int ran_gene_using_genrand_real1(int traject, int wdim,
				 double step, double base_step, double errC,
				 int maxK, double work[], double rani[],
				 double tranij[], double ran_diag[],
				 int *rann_empty, double *rann_rest,
				 double init_acc_rani[]);

int main(void) {
  /* For large memory area, the following variables are allocated
     in static area. */
  static double yvec[NArrayMax], ynew[NArrayMax];
  static double ranni[NWinMax], NormWTend[NWinMax], ranDiag[NWinMax],
#if 1<WDIM
    TransRanIij[NWinWinM1Max]
#else
    TransRanIij[1]
#endif
    ;
  
  unsigned long seed = 5489UL; /* 5489L, 0UL */
  int traject=TRAJECT,ydim=YDIM, wdim=WDIM, stagenum, i, j, ibase,
    rann_empty, ibaseRanni, integ_st_flag;
  double stepleng, sqstep,
    t_point, eps, yinit[YDIM], tmp1, tmp2, sndMom1, sndMom2, errC,
    work_for_ran[(WDIM+2*MaxK+2)*WDIM+(WDIM-1)*WDIM/2], rann_rest;
#if (1 == SRK_TYPE)
  double work[8*(8+3*WDIM)*YDIM];
#elif (2 == SRK_TYPE)
  double work[8*(6+3*WDIM)*YDIM];
#endif

  stagenum=3; /* 3, 41, 45, 53, 58, 104 */

  errC=1.0/2.0;
  
  yinit[0]=yinit[1]=1.0;
  
  stepleng=STEP_SIZE;
  eps = 0.1*stepleng;
  sqstep = sqrt(stepleng);
  
  /* The following sets a seed. */
  init_genrand(seed);
  rann_empty=1;
  
  for(i=ibase=0;i<traject;i++) {/* Initialization for yvec. */
    for(j=0;j<ydim;j++) {
      yvec[ibase+j] = yinit[j];
    }
    ibase+=ydim;
  }
  
  t_point = 0.0; /* Initialization for t_point. */
  
  for(i=ibase=0;i<traject;i++) {/* Initialization for NormWTend. */
    for(j=0;j<wdim;j++) {
      NormWTend[ibase+j]=0;
    }
    ibase+=wdim;
  }
  
  /* t_point keeps changing until it reaches T_END. */
  while(eps<fabs(T_END-t_point)) {
    /* A loop for i to generate random numbers for one step. */
    if(0!=ran_gene_using_genrand_real1(traject, wdim, stepleng,
				       FINE_STEP_SIZE, errC, MaxK,
				       work_for_ran, ranni, TransRanIij,
				       ranDiag, &rann_empty,
				       &rann_rest, NormWTend)) {
      printf("Error in ran_gene_using_genrand_real1!");
      exit(1);
    }
    
    for(i=ibase=0;i<traject;i++) {/* Increment of NormWTend. */
      for(j=0;j<wdim;j++) {
	NormWTend[ibase+j]+=ranni[ibase+j];
      }
      ibase+=wdim;
    }
    
    /* Calculations for ynew by SROCK. */
    if(0 != SRK(ydim, traject, yvec, stepleng, ranni, TransRanIij,
		ranDiag, ffunc, gfunc_gene, wdim, stagenum, work, ynew)) {
      printf("error in SRK\n");
      return 1;
    }
    
    for(i=ibase=0;i<traject;i++) {/* Update on yvec. */
      for(j=0;j<ydim;j++) {
	yvec[ibase+j] = ynew[ibase+j];
      }
      ibase+=ydim;
    }
    t_point += stepleng;           /* Update on t_point. */
  } /* End of loop for t_point */
  
  sndMom1=sndMom2=0;
  for(i=ibase=0;i<traject;i++) {
    j=0;
    sndMom1 += yvec[ibase+j]*yvec[ibase+j];
    j=1;
    sndMom2 += yvec[ibase+j]*yvec[ibase+j];
    ibase+=ydim;
  }
  sndMom1/=traject;
  sndMom2/=traject;

  printf("log2(stepleng) = %3.1lf\n",log(stepleng)/log(2.0));
  printf("2nd mom of y[0] = %16.15lf\n",sndMom1);
  printf("2nd mom of y[1] = %16.15lf\n",sndMom2);
  
  return 0;
}

static void ffunc(double ynvec[],double foutput[])
{
  foutput[0] = A11*ynvec[0];
  foutput[1] = A21*ynvec[0]+A22*ynvec[1];
}

static void gfunc_gene(double ynvec[],int i_th,double goutput[])
{
  switch(i_th) {
  case 0:
    goutput[0] = B11*ynvec[0];
    goutput[1] = B21*ynvec[1];
    break;
  case 1:
    goutput[0] = B12*ynvec[0];
    goutput[1] = B221*ynvec[0]+B222*ynvec[1];
    break;
  default:
    printf("Error_in_gfunc_gene: invalid diffusion coefficient was selected!\n");
    exit(0);
  }
}
